#ifndef LIST_H
#define LIST_H

#include <iostream>

namespace AG{

  template <typename T>
  class list{

  protected:
    struct listNode{   // definition of a single node

      T item_;
      listNode * next_;

      listNode(): next_(0) {};
      listNode(T i, listNode * nx = 0): item_(i), next_(nx) {};
    };

    listNode * head_;   // head of the list
    listNode * tail_;   // tail of the list
    listNode * currentNode_;   // tail of the list
    T defaultValue_;    // default value
      
  private:
 
    // merge
    listNode * merge(listNode * a, listNode * b);

    // mergesort
    listNode * mergesort(listNode* c);

  public:
  
    typedef listNode * link;        // just a shortcut    

    // constructors
    list() : head_(new listNode), 
	     tail_(head_){tail_->next_=0;};
    list (T defaultValue) : 
      head_(new listNode), tail_(head_), 
      defaultValue_(defaultValue){tail_->next_=0;};

    // destructors
    ~list();

    // methods
  
    //    empty
    bool empty() const;

    //    insert
    void insertNode(const T &element);
  
    //    delete
    bool deleteNode(const T &element);

    //    sort
    void sort();

    /////////////////////////////// iterators
    
    struct iter_{

      listNode * it_;

      iter_() {};
      iter_(listNode * it) : it_(it){};

      // increment operator
      const iter_& operator++ (int dummy) 
      { this->it_ = this->it_->next_; return *this; }
      listNode * operator++ () 
      { this->it_ = this->it_->next_; return *this; }

      // dereference operator
      T& operator*() { return this->it_->item_; }
      
      // comparison operator
      bool operator!=(const iter_& other) const { return it_ != other.it_; }; 
    };

    typedef iter_ iterator;             // iterator

    struct iter_const_{

      listNode * it_;

      iter_const_() {};
      iter_const_(listNode * it) : it_(it){};

      // increment operator
      const iter_const_& operator++ (int dummy) 
      { this->it_ = this->it_->next_; return *this; }
      listNode * operator++ () 
      { this->it_ = this->it_->next_; return *this; }

      // dereference operator
      T& operator*() { return this->it_->item_; }
      
      // comparison operator
      bool operator!=(const iter_const_& other) const 
      { return it_ != other.it_; }; 
      bool operator!=(const iter_& other) const 
      { return it_ != other.it_; }; 

      // assignment operator
      iter_const_& operator=(const iter_& other){
	
	if (this != (iter_const_ *) &other)
	  it_ = other.it_;
	return *this;
      }
    };

    typedef iter_const_ const_iterator; // const iterators

    // begin, end, ++ operators
    inline const_iterator begin() const { return const_iter_(head_->next_); }
    inline const_iterator end() const { return const_iter_(tail_->next_); } // which is null
    inline iterator begin(){ return iter_(head_->next_); }
    inline iterator end(){ return iter_(tail_->next_); } // which is null


    /////////////////////////////// iterators

    // friends
    friend std::ostream& operator <<(std::ostream& os, list &l)
    {
      iterator it;
      unsigned int i =1;
      for (it = l.begin(); it !=l.end(); it++)
      	os << i++ << " element = " << *it << std::endl;
      return os;
    }; 

  };

// template definition
#include "list.cpp"

}; // end namespace


#endif
